#import "@local/handout:0.1.0": *
#import "@preview/cetz:0.3.1"
#import "@preview/cetz-plot:0.1.0": plot, chart

= Integers and Floats

#generic("Observation:")
For small values of $a$, $log_2(1 + a)$ is approximately equal to $a$. \
Note that this equality is exact for $a = 0$ and $a = 1$, since $log_2(1) = 0$ and $log_2(2) = 1$.

#v(2mm)

We'll add a "correction term" $epsilon$ to this approximation, so that $log_2(1 + a) approx a + epsilon$.

#cetz.canvas({
  import cetz.draw: *

  let f1(x) = calc.log(calc.abs(x + 1), base: 2)
  let f2(x) = x

  // Set-up a thin axis style
  set-style(axes: (stroke: .5pt, tick: (stroke: .5pt)))


  plot.plot(
    size: (8, 8),
    x-tick-step: 0.2,
    y-tick-step: 0.2,
    y-min: 0,
    y-max: 1,
    x-min: 0,
    x-max: 1,
    legend: none,
    axis-style: "scientific-auto",

    {
      let domain = (0, 10)

      plot.add-fill-between(
        f1,
        f2,
        domain: domain,
        style: (stroke: none, fill: luma(75%)),
      )

      plot.add(
        f1,
        domain: domain,
        label: $log(1+x)$,
        style: (stroke: black),
      )
      plot.add(f2, domain: domain, label: $x$, style: (stroke: black))
    },
  )
})

TODO: why? Graphs.

#problem(label: "convert")
Use the fact that $log_2(1 + a) approx a + epsilon$ to approximate $log_2(x_f)$ in terms of $x_i$. \

#v(5mm)

Namely, show that
$
  log_2(x_f) = (x_i) / (2^23) - 127 + epsilon
$
for some correction term term $epsilon$ \
#note([
  In other words, we're finding an expression for $x$ as a float
  in terms of $x$ as an int.
])

#solution([
  Let $E$ and $F$ be the exponent and float bits of $x_f$. \
  We then have:
  $
    log_2(x_f)
    &= log_2 ( 2^(E-127) times (1 + (F) / (2^23)) ) \
    &= E - 127 + log_2(1 + F / (2^23)) \
    & approx E-127 + F / (2^23) + epsilon \
    &= 1 / (2^23)(2^23 E + F) - 127 + epsilon \
    &= 1 / (2^23)(x_i) - 127 + epsilon
  $
])

#v(1fr)
