#import "@local/handout:0.1.0": *

= The Fast Inverse Square Root


The following code is present in _Quake III Arena_ (1999):
```c
float Q_rsqrt( float number ) {
    long i = * ( long * ) &number;
    i = 0x5f3759df - ( i >> 1 );
    return * ( float * ) &i;
}
```

This code defines a function `Q_rsqrt` that consumes a float `number` and approximates its inverse square root.
If we rewrite this using notation we're familiar with, we get the following:
$
  #text[`Q_sqrt`] (n_f) = 6240089 - (n_i div 2)
  approx 1 / sqrt(n_f)
$

#note([
  `0x5f3759df` is $6240089$ in hexadecimal. \
  It is a magic number hard-coded into `Q_sqrt`.
])

#v(2mm)

Our goal in this section is to understand why this works:
- How does Quake approximate $1 / sqrt(x)$ by simply subtracting and dividing by two?
- What's special about $6240089$?


#problem()
Using basic log rules, rewrite $log_2(1 / sqrt(x))$ in terms of $log_2(x)$.

#solution([
  $
    log_2(1 / sqrt(x)) = (-1) / (2)log_2(x)
  $
])

#v(1fr)
#pagebreak()

#generic("Setup:")
We are now ready to show that $#text[`Q_sqrt`] (x) approx 1/sqrt(x)$. \
For convenience, let's call the bit string of the inverse square root $r$. \
In other words,
$
  r_f := 1 / (sqrt(n_f))
$
This is the value we want to approximate.

#problem(label: "finala")
Find an approximation for $log_2(r_f)$ in terms of $n_i$ and $epsilon$ \
#note[Remember, $epsilon$ is the correction constant in our approximation of $log_2(1 + a)$.]

#solution[
  $
    log_2(r_f)
    = log_2(1 / sqrt(n_f))
    = (-1) / 2 log_2(n_f)
    approx (-1) / 2 ( (n_i) / (2^23) + epsilon - 127 )
  $
]

#v(1fr)

#problem(label: "finalb")
Let's call the "magic number" in the code above $kappa$, so that
$
  #text[`Q_sqrt`] (n_f) = kappa - (n_i div 2)
$
Use @convert and @finala to show that $#text[`Q_sqrt`] (n_f) approx r_i$

#solution[
  From @convert, we know that
  $
    log_2(r_f) approx (r_i) / (2^23) + epsilon - 127
  $

  #note[
    Our approximation of $log_2(1+a)$ uses a fixed correction constant, \
    so the $epsilon$ here is equivalent to the $epsilon$ in @finala.
  ]

  Combining this with the result from \ref{finala}, we get:
  $
    (r_i) / (2^23) + epsilon - 127
    &approx (-1) / (2) ( (n_i) / (2^23) + epsilon - 127) \
    (r_i) / (2^23)
    &approx (-1) / (2) ( (n_i) / (2^23)) + 3 / 2 (epsilon - 127) \
    r_i
    &approx (-1) / 2 (n_i) + 2^23 3 / 2(epsilon - 127)
    = 2^23 3 / 2 (epsilon - 127) - (n_i) / 2
  $

  #v(2mm)

  This is exactly what we need! If we set $kappa$ to $(2^24)/3 (epsilon - 127)$, then
  $
    r_i approx kappa - (n_i div 2) = #text[`Q_sqrt`] (n_f)
  $
]

#v(1fr)

#problem()
What is the exact value of $kappa$ in terms of $epsilon$? \
#hint[Look at @finalb. We already found it!]

#solution[
  This problem makes sure our students see that
  $kappa = (2^24)/3(epsilon - 127)$. \
  See the solution to @finalb.
]

#if_no_solutions(v(2cm))
